
//	Copyright (c) 1982 Dyadic Systems Limited	

/*	x f a c c

	external function access routines
*/

#include <apl.h>
#include <support.h>

static int32_t	roundi(double fval);

//	d o u b l e   p u t  &  g e t	

typedef	struct {
	UCH cfloat[sizeof(aplfloat64)];
} Doubarray;

typedef struct {
	union {
		Doubarray sfloat;
		aplfloat64 dfloat;
	} ufloat;
} Doubunion;

void putd(aplfloat64 val, aplfloat64 *base, BOUND indx)
{
#ifdef	DALIGN
	Doubunion dunion;

	dunion.ufloat.dfloat = val;
	*((Doubarray*)(base+indx)) = dunion.ufloat.sfloat;
#else
	base[indx] = val;
#endif // DALIGN 
}

aplfloat64 getd(void *vbase, BOUND indx)
{
	aplfloat64 *base = vbase;
#ifdef	DALIGN
	Doubunion dunion;

	dunion.ufloat.sfloat = *((Doubarray*)(base+indx));
	return dunion.ufloat.dfloat;
#else

	return base[indx];

#endif // DALIGN 
}
//	c h a r  f r o m  a n y t h i n g  ( s u i t a b l e )	

static aplchar8 cgetc(void *vbase, BOUND indx)
{
	aplchar8 *base = vbase;

	return base[indx];
}

static aplchar8 cgetp(void *vbase, BOUND indx)
{
	ARRAY **base = vbase;
	ARRAY *array = base[indx];

	if			// subarray should be simple scalar
		(
		array->eltype == APLPNTR
	||	array->rank
		) domain();

	return (*initchar(array))(array->shape, 0);
}

static aplchar8 cdomx(void *vbase, BOUND indx)
	{
	domain();
	return 0;	// keep compiler happy
	}

//	i n t   f r o m   a n y t h i n g	

static aplint16 igetb(void *vbase, BOUND indx)
{
	aplint16 *base = vbase;

	return ((base[indx>>3] << (indx&7)) & 0x80 ) != 0;
}

static aplint16 igets(void *vbase, BOUND indx)
{
	aplint8 *base = vbase;

	return base[indx];
}

static aplint16 igeti(void *vbase, BOUND indx)
{
	aplint16 *base = vbase;

	return base[indx];
}

static aplint16 igetl(void *vbase, BOUND indx)
{
	aplint32 *base = vbase;

	return base[indx];
}

static aplint16 igetd(void *vbase, BOUND indx)	// rounds double to nearest int 
{
	aplfloat64 *base = vbase;
#ifdef	DALIGN
	Doubunion dunion;

	dunion.ufloat.sfloat = *((Doubarray*)(base+indx));
	return roundi(dunion.ufloat.dfloat);
#else
	return roundi(base[indx]);
#endif // DALIGN 
}

static aplint16 igetp(void *vbase, BOUND indx)
{
	ARRAY **base = vbase;
	ARRAY *array = base[indx];

	if			// subarray should be simple scalar
		(
		array->eltype == APLPNTR
	||	array->rank
		) domain();

	return (*initintg(array))(array->shape, 0);
}

static aplint16 idomx(void *vbase, BOUND indx)
	{
	domain();
	return 0; // keep compiler happy
	}

//	l o n g   f r o m   a n y t h i n g	

static aplint32 lgetb(void *vbase, BOUND indx)
{
	apluns8 *base = vbase;

	return ((base[indx>>3] << (indx&7)) & 0x80 ) != 0;
}

static aplint32 lgets(void *vbase, BOUND indx)
{
	aplint8 *base = vbase;

	return base[indx];
}

static aplint32 lgeti(void *vbase, BOUND indx)
{
	aplint16 *base = vbase;

	return base[indx];
}

static aplint32 lgetl(void *vbase, BOUND indx)
{
	aplint32 *base = vbase;

	return base[indx];
}

static aplint32 lgetd(void *vbase, BOUND indx)	// rounds double to nearest long
{
	aplfloat64 *base = vbase;
#ifdef	DALIGN
	Doubunion dunion;

	dunion.ufloat.sfloat = *((Doubarray*)(base+indx));
	return roundi(dunion.ufloat.dfloat);
#else
	return roundi(base[indx]);
#endif // DALIGN 
}

static aplint32 lgetp(void *vbase, BOUND indx)
{
	ARRAY **base = vbase;
	ARRAY *array = base[indx];

	if			// subarray should be simple scalar
		(
		array->eltype == APLPNTR
	||	array->rank
		) domain();

	return (*initlong(array))(array->shape, 0);
}

static aplint32 ldomx(void *vbase, BOUND indx)
	{
	domain();
	return 0;	// keep compiler happy
	}

//	d o u b l e   f r o m   a n y t h i n g	

static aplfloat64 dgetb(void *vbase, BOUND indx)
{
	apluns8 *base = vbase;

	return ((base[indx>>3] << (indx&7)) & 0x80 ) != 0;
}

static aplfloat64 dgets(void *vbase, BOUND indx)
{
	aplint8 *base = vbase;

	return  (int)base[indx];
}

static aplfloat64 dgeti(void *vbase, BOUND indx)
{
	aplint16 *base = vbase;

	return base[indx];
}

static aplfloat64 dgetl(void *vbase, BOUND indx)
{
	aplint32 *base = vbase;

	return base[indx];
}

static aplfloat64 dgetp(void *vbase, BOUND indx)
{
	ARRAY **base = vbase;
	register ARRAY *array = base[indx];

	if			// subarray should be simple scalar
		(
		array->eltype == APLPNTR
	||	array->rank
		) domain();
	return (*initdoub(array))(array->shape, 0);
}

static aplfloat64 ddomx(void *vbase, BOUND indx)
	{
	domain();
	return 0;	// keep compiler happy
	}

//	p n t r  f r o m  a n y t h i n g	

static ARRAY *pgetc(void *vbase, BOUND indx)
{
	aplchar8 *base = vbase;
	register ARRAY *array;

	array = mkarray(APLNCHAR, 0, 0);

	*(aplchar8 *)array->shape = base[indx];

	return array;
}

static ARRAY *pgetb(void *vbase, BOUND indx)
{
	apluns8 *base = vbase;
	register ARRAY *array;

	array = mkarray(APLLONG, 0, 0);

	*(aplint32 *)array->shape = lgetb(base,indx);

	return array;
}

static ARRAY *pgets(void *vbase, BOUND indx)
{
	aplint8 *base = vbase;
	ARRAY *array;

	array = mkarray(APLLONG, 0, 0);

	*(aplint32 *)array->shape = base[indx];

	return array;
}

static ARRAY *pgeti(void *vbase, BOUND indx)
{
	aplint16 *base = vbase;
	register ARRAY *array;

	array = mkarray(APLLONG, 0, 0);

	*(aplint32 *)array->shape = base[indx];

	return array;
}

static ARRAY *pgetl(void *vbase, BOUND indx)
{
	aplint32 *base = vbase;
	ARRAY *array;

	array = mkarray(APLLONG, 0, 0);

	*(aplint32 *)array->shape = base[indx];

	return array;
}

static ARRAY *pgetd(void *vbase, BOUND indx)
{
	aplfloat64 *base = vbase;
	ARRAY *array;

	array = mkarray(APLDOUB, 0, 0);

	*(aplfloat64 *)array->shape = base[indx];

	return array;
}

static ARRAY *pgetp(void *vbase, BOUND indx)
{
	ARRAY **base = vbase;

	return base[indx];
}

static ARRAY *pdomx(void *vbase, BOUND indx)
	{
	domain();
	}

//	S E R V I C E S		

static aplchar8 (*charfns[])(void *vbase, BOUND indx) = 
{
			cgetc,
	cdomx,				// APLBOOL 
	cdomx,				// APLSINT 
	cdomx,				// APLINTG 
	cdomx,				// APLLONG 
	cdomx,				// APLDOUB 
			cgetp
};

static aplint16 (*shortfns[])(void *vbase, BOUND indx) =
{
	idomx,				// APLNCHAR	
	igetb, igets, igeti, igetl, igetd, igetp
};

static aplint32 (*longfns[])(void *vbase, BOUND indx) =
{
	ldomx,				// APLNCHAR	
	lgetb, lgets, lgeti, lgetl, lgetd, lgetp
};

static aplfloat64 (*doublefns[])(void *vbase, BOUND indx) = 
{
 	ddomx,				// APLNCHAR 
	dgetb, dgets, dgeti, dgetl, getd, dgetp
};

static ARRAY	*(*pntrfns[])(void *vbase, BOUND indx) = 
{
	pgetc, pgetb, pgets, pgeti, pgetl, pgetd, pgetp
};

aplchar8 (*initchar(ARRAY *array))(void *vbase, BOUND indx)
{
	return charfns[array->eltype];
}

aplint16 (*initintg(ARRAY *array))(void *vbase, BOUND indx)
{
	return shortfns[array->eltype];
}

aplint32 (*initdint(ARRAY *array))(void *vbase, BOUND indx)
{
	return (sizeof(int) == sizeof(short)
			? (int (**)(void *, BOUND))shortfns
			: (int (**)(void *, BOUND))longfns)[array->eltype];
}

aplint32 (*initlong(ARRAY *array))(void *vbase, BOUND indx)
{
	return longfns[array->eltype];
}

aplfloat64 (*initdoub(ARRAY *array))(void *vbase, BOUND indx)
{
	return doublefns[array->eltype];
}

ARRAY	*(*initpntr(ARRAY *array))(void *vbase, BOUND indx)
{
	return pntrfns[array->eltype];
}

static aplint32 roundi(double fval)	// round to integer
	{
	static double max = 0x7fffffff;
	static double min = (double)(-1) - (double) 0x7fffffff;

	fval += (fval<0 ? -.5:.5);
	if(min > fval || max < fval)
		domain();
	return fval;
	}

//	Copyright (c) 1982 Dyadic Systems Limited	
