//	m k a r r a y	

#include <apl.h>
#include <stdlib.h>
#include <string.h>

static void (*array_err_fn)(int);

void install_array_err_fn(void (*fn)(int))
{
array_err_fn=fn;
}

static ARRAY *array_err(int err)
{
if (array_err_fn)(*array_err_fn)(err);
return (ARRAY *)0;
}

ARRAY *alloc(PLENGTH nwds)			// similar to C library 'malloc'	
{
	UCH *p;

	if((p = malloc(nwds*sizeof(MEMORYINT))) == 0)
		array_err(LIMIT);

	return (ARRAY *)p;
}

static UCH SHIFT[] =	// to convert no. of elements to no. of bits 
{	3, 	// APLCHAR 
	0, 	// APLBOOL 
	3, 	// APLSINT 
	4, 	// APLINTG 
	5, 	// APLLONG 
	6, 	// APLDOUB 
	PNTRlog,// APLPNTR 
	3, 	// APLWCHAR8 
	4, 	// APLWCHAR16 
	5 	// APLWCHAR32 
};

ARRAY *mkarray
	(					// make an array in memory 
	unsigned eltype,			// element type 
	unsigned rank,				// array rank 
	BOUND *shape				// pointer to shape vector 
	)
	{
	BOUND bound = 1, nwds, *ptr, was;
	unsigned u;
	ARRAY *array;
	unsigned overflow = 0;
	int ishift;

	if(eltype >= NELTYPES)
		return array_err(EDOMAIN);
	if(rank   >  MAXRANK )
		return array_err(RANK);

	for(ptr = shape, u = rank; u--; ++ptr)	// calc no of elements as 
		{
		BOUND this = *ptr;

		if(this)
			{
			was = bound;
			bound *= *ptr;		// product of shape words 
			overflow |= was != (bound / *ptr);
			}
		else				// special case zero
			{
			overflow = 0;
			bound = 0;
			break;
			}
		}

	if (overflow)
		return array_err(LIMIT);

	ishift = SHIFT[eltype] - PNTRlog;
	if(0 <= ishift)
		{
		static BOUND maxima[] =
			{
#if	5 == PNTRlog
			0x80000000
		,	0xc0000000
		,	0xe0000000
		,	0xf0000000
		,	0xf8000000
		,	0xfc000000
#elif	6 == PNTRlog
			0x8000000000000000
		,	0xc000000000000000
		,	0xe000000000000000
		,	0xf000000000000000
		,	0xf800000000000000
		,	0xfc00000000000000
#else
#	error	extra code needed here
#endif//PNTRlog
			};

		if(bound & maxima[ishift])
			return array_err(LIMIT);
		nwds = bound << ishift;		// nwds >= bound
		}
	else
		{
		ishift = -ishift;
		nwds = bound >> ishift;
		nwds += 0 != (bound - (nwds << ishift)); // round up if req
		}
	was = nwds;
	nwds +=					// add extras
		2				// length and zones
		+ rank				// shape vector		
		+ (eltype == APLPNTR && bound == 0);	// prototype if null nested 
	if (was > nwds)				// overflow
		return array_err(LIMIT);

#ifdef	DALIGN
	if(eltype == APLDOUB)
		{
		MEMORYUNS *tmp;
		unsigned pad;

		nwds += pad = (rank+2)%DOUBWORDS;
		if (was > nwds)
			return array_err(LIMIT);
		tmp = (MEMORYUNS *)alloc(nwds);
		nwds -= pad;
		array = (ARRAY *)(tmp+(DOUBWORDS-pad)%DOUBWORDS);
		}
	else
#endif	// DALIGN 
	array = alloc(nwds);			// allocate space for array 
						// and fill in type ...	
	if(eltype == APLPNTR)			// zero so freearray() works 
		memset(array, 0, nwds * sizeof(MEMORYUNS));
	array->type = (eltype == APLPNTR ? NESTED : SIMPLE);
	array->length = nwds;			// length		
	array->eltype = eltype;			// element type		
	array->rank = rank;			// rank			

	memcpy(array->shape, shape, rank * sizeof(MEMORYUNS));

	return array;
	}

/*	a r r a y b o u n d
	return number of elements in an array
	null nested arrays have one element anyway
*/

BOUND	arraybound(ARRAY *array)
{
	BOUND	bd, n, *p;

	bd = 1;
	p = array->shape;

	for(n = array->rank; n--; ++p)
		bd *= *p;

	if(array->eltype == APLPNTR && !bd) bd = 1;
	return bd;
}

/*	a r r a y s t a r t
	return pointer to first byte of data portion
*/

void *arraystart(ARRAY *array)
{
	return	array->shape + array->rank;
}

